import io
import json
import typing
from datetime import datetime
from uuid import uuid4
from string import hexdigits
from collections import defaultdict, OrderedDict

from botocore.exceptions import ClientError

from flashflood.util import datetime_from_timestamp, timestamp_now, S3Deleter
from flashflood.exceptions import FlashFloodException, FlashFloodEventNotFound, FlashFloodJournalUploadError
from flashflood.identifiers import JournalID, JournalUpdateID, JournalUpdateAction, TOMBSTONE_SUFFIX


class Event(typing.NamedTuple):
    event_id: str
    date: datetime
    data: bytes


class BaseJournalUpdate:
    bucket: typing.Any = None
    _pfx: typing.Optional[str] = None

    def __init__(self, update_id: JournalUpdateID):
        self.id_ = update_id
        self._data: typing.Optional[bytes] = None

    @classmethod
    def make(cls, journal_id: JournalID, event_id: str, action: JournalUpdateAction):
        id_ = JournalUpdateID.make(journal_id, event_id, action)
        return cls(JournalUpdateID(id_))

    @classmethod
    def from_key(cls, key: str):
        return cls(JournalUpdateID.from_key(key))

    @classmethod
    def upload_update(cls, journal_id: JournalID, event_id: str, data: bytes):
        id_ = JournalUpdateID.make(journal_id, event_id, JournalUpdateAction.UPDATE)
        return cls(id_)._upload(data)

    @classmethod
    def upload_delete(cls, journal_id: JournalID, event_id: str):
        id_ = JournalUpdateID.make(journal_id, event_id, JournalUpdateAction.DELETE)
        return cls(id_)._upload(b"")

    def _upload(self, data: bytes):
        key = f"{self._pfx}/{self.id_}"
        self.bucket.Object(key).upload_fileobj(io.BytesIO(data))

    @property
    def journal_id(self) -> JournalID:
        return self.id_.journal_id

    @property
    def event_id(self) -> str:
        return self.id_.event_id

    @property
    def action(self) -> JournalUpdateAction:
        return self.id_.action

    @property
    def data(self) -> bytes:
        if not self._data:
            key = f"{self._pfx}/{self.id_}"
            try:
                self._data = self.bucket.Object(key).get()['Body'].read()
            except ClientError as ex:
                if ex.response['Error']['Code'] == "NoSuchKey":
                    raise FlashFloodException(f"JournalUpdate not found for key {key}")
                raise
        if self._data is None:
            raise FlashFloodException(f"Update {self.id_} has no data!")
        else:
            return self._data

    def delete(self):
        key = f"{self._pfx}/{self.id_}"
        try:
            next(iter(self.bucket.objects.filter(Prefix=key)))
        except StopIteration:
            raise FlashFloodException(f"Cannot delete non-existent object {key}")
        self.bucket.Object(key=f"{key}{TOMBSTONE_SUFFIX}").upload_fileobj(io.BytesIO(b""))

    @classmethod
    def list_out_of_date_journals(cls) -> typing.Iterator[JournalID]:
        prev_journal_id = str()
        for update_id in cls.list():
            if prev_journal_id != update_id.journal_id:
                yield update_id.journal_id
            prev_journal_id = update_id.journal_id

    @classmethod
    def get_updates_for_all_journals(cls) -> typing.Iterator[typing.Tuple[JournalID, typing.Mapping[str, typing.Any]]]:
        curr_journal_id = JournalID()
        updates: typing.Dict[str, typing.Any] = dict()
        for id_ in cls.list():
            if curr_journal_id != id_.journal_id:
                if updates:
                    yield curr_journal_id, updates
                curr_journal_id = id_.journal_id
                updates = {id_.event_id: cls(id_)}
            else:
                updates[id_.event_id] = cls(id_)
        if updates:
            yield curr_journal_id, updates

    @classmethod
    def get_updates_for_journal(cls, journal_id: JournalID) -> dict:
        update_id_pfx = JournalUpdateID.prefix_for_journal(journal_id)
        updates = dict()
        for id_ in cls.list(update_id_pfx):
            update = cls(id_)
            updates[update.event_id] = update
        return updates

    @classmethod
    def list(cls, update_id_pfx="") -> typing.Iterator[JournalUpdateID]:
        prev_key = None
        key = None
        for item in cls.bucket.objects.filter(Prefix=f"{cls._pfx}/{update_id_pfx}"):
            key = item.key
            if prev_key:
                if not key.endswith(TOMBSTONE_SUFFIX):
                    if not prev_key.endswith(TOMBSTONE_SUFFIX):
                        yield JournalUpdateID.from_key(prev_key)
            prev_key = key
        if key and not key.endswith(TOMBSTONE_SUFFIX):
            yield JournalUpdateID.from_key(key)


class BaseJournal:
    bucket: typing.Any = None
    _journal_pfx: typing.Optional[str] = None
    _blobs_pfx: typing.Optional[str] = None

    def __init__(self, events: list=None, blob_id: str=None, data: bytes=None, version: str=None):
        self.events = events or list()
        self.blob_id = blob_id or str(uuid4())
        self.data = data or b""
        self._body: typing.Optional[typing.BinaryIO] = None
        self._location = "memory"
        self.version = version or timestamp_now()

    @classmethod
    def from_key(cls, key: str):
        id_ = JournalID.from_key(key)
        blob_id = id_.blob_id
        try:
            manifest = json.loads(cls.bucket.Object(key).get()['Body'].read().decode("utf-8"))
        except ClientError as ex:
            if ex.response['Error']['Code'] == "NoSuchKey":
                raise FlashFloodException(f"Journal not found for key {key}")
            else:
                raise
        except json.decoder.JSONDecodeError:
            print("Unable to decode manifest document from key:", key)
            raise
        journal = cls(manifest['events'], blob_id, version=id_.version)
        journal._location = "cloud"
        return journal

    @classmethod
    def from_id(cls, journal_id: JournalID):
        return cls.from_key(f"{cls._journal_pfx}/{journal_id}")

    @property
    def body(self) -> typing.BinaryIO:
        if self._body is None:
            if "memory" == self._location:
                self._body = io.BytesIO(self.data)
            elif "cloud" == self._location:
                key = f"{self._blobs_pfx}/{self.blob_id}"
                self._body = self.bucket.Object(key).get()['Body']
            else:
                raise ValueError(f"Unknown data location {self._location}")
        return self._body

    def reload(self):
        self._body = None

    @property
    def is_empty(self) -> bool:
        return 0 == len(self.events)

    @property
    def id_(self) -> JournalID:
        if self.is_empty:
            raise FlashFloodException("Cannot generate id for empty journal")
        else:
            from_timestamp = self.events[0]['timestamp']
            to_timestamp = self.events[-1]['timestamp']
            return JournalID.make(from_timestamp, to_timestamp, self.version, self.blob_id)

    @property
    def size(self):
        if "memory" == self._location:
            return len(self.data)
        elif "cloud" == self._location:
            return sum(e['size'] for e in self.events)
        else:
            raise ValueError(f"Unknown data location {self._location}")

    def manifest(self) -> dict:
        return dict(journal_id=self.id_,
                    from_date=self.events[0]['timestamp'],
                    to_date=self.events[-1]['timestamp'],
                    size=self.size,
                    events=self.events)

    def get_event(self, event_id: str) -> Event:
        for e in self.events:
            if e['event_id'] == event_id:
                blob_key = f"{self._blobs_pfx}/{self.blob_id}"
                byte_range = f"bytes={e['offset']}-{e['offset'] + e['size'] - 1}"
                data = self.bucket.Object(blob_key).get(Range=byte_range)['Body'].read()
                return Event(event_id, datetime_from_timestamp(e['timestamp']), data)
        raise FlashFloodEventNotFound(f"Event {event_id} not found in journal {self.id_}")

    def updated(self, updates: typing.Mapping[str, BaseJournalUpdate]):
        if not updates:
            return self
        else:
            self.reload()
            new_journal_data = b""
            new_events = list()
            for e in self.events:
                event_data = self.body.read(e['size'])
                e['offset'] = len(new_journal_data)
                update = updates.get(e['event_id'], None)
                if update is None:
                    new_journal_data += event_data
                    new_events.append(e)
                elif JournalUpdateAction.UPDATE == update.action:
                    new_journal_data += update.data
                    e['size'] = len(update.data)
                    new_events.append(e)
                elif JournalUpdateAction.DELETE == update.action:
                    pass
                else:
                    raise Exception("No handler for journal update {update}")
            return type(self)(new_events, data=new_journal_data)

    def upload(self) -> str:
        if self.events:
            manifest = self.manifest()
            blob_key = f"{self._blobs_pfx}/{self.blob_id}"
            self.bucket.Object(blob_key).upload_fileobj(self.body, ExtraArgs=dict(Metadata=dict(journal_id=self.id_)))
            key = f"{self._journal_pfx}/{self.id_}"
            metadata = dict(number_of_events=f"{len(self.events)}",
                            journal_data_size=f"{len(self.data)}")
            self.bucket.Object(key).upload_fileobj(io.BytesIO(json.dumps(manifest).encode("utf-8")),
                                                   ExtraArgs=dict(Metadata=metadata))
            self.reload()  # make self._body available to for read again
            print("Uploaded journal", self.id_)
        else:
            raise FlashFloodJournalUploadError("Cannot upload journal with no events")
        return key

    def delete(self):
        key = f"{self._journal_pfx}/{self.id_}"
        try:
            next(iter(self.bucket.objects.filter(Prefix=key)))
        except StopIteration:
            raise FlashFloodException(f"Cannot delete non-existent object {key}")
        self.bucket.Object(key=f"{key}{TOMBSTONE_SUFFIX}").upload_fileobj(io.BytesIO(b""))

    def keys(self):
        id_ = self.id_
        return [f"{self._journal_pfx}/{id_}", f"{self._blobs_pfx}/{id_.blob_id}"]

    @classmethod
    def list(cls) -> typing.Iterator[JournalID]:
        # TODO: heuristic to find from_date in bucket listing -xbrianh
        journal_info: dict = dict(range_prefix=None, journal_ids=list())
        for item in cls.bucket.objects.filter(Prefix=cls._journal_pfx):
            journal_id = JournalID.from_key(item.key)
            if journal_id.range_prefix != journal_info['range_prefix']:
                if journal_info['journal_ids']:
                    yield journal_info['journal_ids'][-1]
                journal_info['range_prefix'] = journal_id.range_prefix
                journal_info['journal_ids'] = [journal_id]
            else:
                if journal_id.endswith(TOMBSTONE_SUFFIX):
                    journal_info['journal_ids'].remove(journal_id.replace(TOMBSTONE_SUFFIX, ""))
                else:
                    journal_info['journal_ids'].append(journal_id)
        if journal_info['journal_ids']:
            yield journal_info['journal_ids'][-1]



import io
import typing

from flashflood.util import delete_keys

class BaseKeyIndex:
    """
    Build a simple key value index using s3 keys. Updates are modeled as writes to avoid S3 eventual consistency for
    overwrites.

    Concurrent writes are not supported.
    """
    DELIMITER: str = "--"
    bucket: typing.Any = None
    _pfx: typing.Optional[str] = None

    @classmethod
    def put(cls, lookup: str, target: str):
        keys = cls._put(lookup, target)
        delete_keys(cls.bucket, keys)

    @classmethod
    def put_batch(cls, lookup_map: dict):
        keys_to_delete = list()
        for lookup, target in lookup_map.items():
            keys = cls._put(lookup, target)
            keys_to_delete.extend(keys)
        delete_keys(cls.bucket, keys_to_delete)

    @classmethod
    def _put(cls, lookup: str, target: str) -> list:
        keys = cls._lookup_keys(lookup)
        if keys:
            revision_number = cls._revision_number_for_key(keys[-1]) + 1
        else:
            revision_number = 1
        revision = "%010i" % revision_number
        key = f"{cls._pfx}/{lookup}" + cls.DELIMITER + revision
        cls.bucket.Object(key).upload_fileobj(io.BytesIO(b""),
                                              ExtraArgs=dict(Metadata=dict(target=target)))
        return keys

    @classmethod
    def delete(cls, lookup: str):
        keys = cls._lookup_keys(lookup)
        if keys:
            delete_keys(cls.bucket, keys)

    @classmethod
    def get(cls, lookup: str):
        keys = cls._lookup_keys(lookup)
        if keys:
            return cls.bucket.Object(keys[-1]).metadata['target']
        else:
            return None

    @classmethod
    def _lookup_keys(cls, lookup: str):
        return [item.key for item in cls.bucket.objects.filter(Prefix=f"{cls._pfx}/{lookup}")]

    @classmethod
    def _revision_number_for_key(cls, key: str) -> int:
        return int(key.rsplit(cls.DELIMITER, 1)[1])



class FlashFloodException(Exception):
    pass

class FlashFloodJournalingError(FlashFloodException):
    pass

class FlashFloodEventExistsError(FlashFloodException):
    pass

class FlashFloodEventNotFound(FlashFloodException):
    pass

class FlashFloodJournalUploadError(FlashFloodException):
    pass



import io
from datetime import datetime
import json
import typing
import requests
from uuid import uuid4
from concurrent.futures import ThreadPoolExecutor, as_completed

from flashflood.util import datetime_to_timestamp, datetime_from_timestamp, DateRange, S3Deleter
from flashflood.objects import Event, BaseJournal, BaseJournalUpdate
from flashflood.identifiers import JournalID, TOMBSTONE_SUFFIX
from flashflood.key_index import BaseKeyIndex
from flashflood.exceptions import (FlashFloodException, FlashFloodEventNotFound, FlashFloodEventExistsError,
                                   FlashFloodJournalingError)


# TODO:
# What happens if event updates are written for a journal that has been removed?
#
# It's possible to resolve journals for update markers that are out-of-date using the event index
#
# Event updates should be applied idempotently
#
# write tests for multiple updates/deletes to the same event
# Create resolution policy for multiple updates/deletes to the same event
#
# Use bucket lifecycle policy for garbage collection
# Warn if policy is not set upon FF instantiation
#
# Include size/len(events) metadat in journal listing?


class FlashFlood:
    def __init__(self, s3_resource: typing.Any, bucket: str, root_prefix: str):
        self.s3 = s3_resource
        self.bucket = self.s3.Bucket(bucket)
        if root_prefix.endswith("/"):
            raise ValueError("Root prefix cannot end with `/`")
        self.root_prefix = root_prefix
        self._journal_pfx = f"{root_prefix}/journals"
        self._blobs_pfx = f"{root_prefix}/blobs"
        self._update_pfx = f"{root_prefix}/update"
        self._index_pfx = f"{root_prefix}/index"

        class _Journal(BaseJournal):
            bucket = self.bucket
            _journal_pfx = self._journal_pfx
            _blobs_pfx = self._blobs_pfx

        class _JournalUpdate(BaseJournalUpdate):
            bucket = self.bucket
            _pfx = self._update_pfx

        class _KeyIndex(BaseKeyIndex):
            bucket = self.bucket
            _pfx = self._index_pfx

        self._Journal = _Journal
        self._JournalUpdate = _JournalUpdate
        self._KeyIndex = _KeyIndex

    def put(self, data, event_id: str=None, date: datetime=None) -> Event:
        date = date or datetime.utcnow()
        timestamp = datetime_to_timestamp(date)
        event_id = event_id or str(uuid4())
        if JournalID.DELIMITER in event_id:
            raise FlashFloodException(f"'{JournalID.DELIMITER}' not allowed in event_id")
        if self.event_exists(event_id):
            raise FlashFloodEventExistsError(f"Event {event_id} already exists")
        else:
            events = [dict(event_id=event_id, timestamp=timestamp, offset=0, size=len(data))]
            journal = self._Journal(events, data=data, version="new")
            journal.upload()
            self._index_journal(journal)
            print("new journal", journal.id_)
            return Event(event_id, date, data)

    def _index_journal(self, journal: BaseJournal):
        journal_id = journal.id_
        self._KeyIndex.put_batch({e['event_id']: journal_id for e in journal.events})

    def update_event(self, event_id: str, new_data: bytes):
        if self.event_exists(event_id):
            journal_id = self._journal_for_event(event_id)
            self._JournalUpdate.upload_update(journal_id, event_id, new_data)
        else:
            raise FlashFloodEventNotFound(f"Event {event_id} not found")

    def delete_event(self, event_id: str):
        """
        Write delete marker for event.
        """
        journal_id = self._journal_for_event(event_id)
        self._JournalUpdate.upload_delete(journal_id, event_id)
        # Deindexing immediately means the event is not available for lookup, but it will still appear in replay.
        self._KeyIndex.delete(event_id)

    def update(self, number_of_updates_to_apply: int=1000) -> int:
        """
        Apply updates for each update and delete marker.
        """
        count = 0
        for journal_id, updates in self._JournalUpdate.get_updates_for_all_journals():
            print("updating journal", journal_id)
            journal = self._Journal.from_id(journal_id)
            new_journal = journal.updated(updates)
            if new_journal != journal:
                if not new_journal.is_empty:
                    new_journal.upload()
                    self._index_journal(new_journal)
                journal.delete()
            for u in updates.values():
                u.delete()
            count += len(updates)
            if number_of_updates_to_apply <= count:
                break
        return count

    def journal(self, minimum_number_of_events: int=100, minimum_size: int=None):
        minimum_size = minimum_size or 0
        number_of_events, size = 0, 0
        journals_to_combine = list()
        for journal_id in self._new_journals():
            journal = self._Journal.from_id(journal_id)
            size += journal.size
            number_of_events += len(journal.events)
            journals_to_combine.append(journal)
            print("Found journal to combine", journal.id_)
            if minimum_number_of_events <= number_of_events and minimum_size <= size:
                break
        if minimum_number_of_events > number_of_events:
            raise FlashFloodJournalingError(f"Journal condition: minimum_number_of_events={minimum_number_of_events}")
        if minimum_size > size:
            raise FlashFloodJournalingError(f"Journal condition: minimum_size={minimum_size}")
        self.combine_journals(journals_to_combine)

    def _new_journals(self) -> typing.Iterator[JournalID]:
        for journal_id in self._Journal.list():
            if "new" == journal_id.version:
                yield journal_id

    def combine_journals(self, journals_to_combine: typing.List[BaseJournal]) -> BaseJournal:
        new_journal = self._Journal()
        objects_to_delete = journals_to_combine.copy()
        for journal in journals_to_combine:
            print("combining journal", journal.id_)
            updates = self._JournalUpdate.get_updates_for_journal(journal.id_)
            objects_to_delete.extend(list(updates.values()))
            journal = journal.updated(updates)
            for e in journal.events:
                new_offset = e['offset'] + len(new_journal.data)
                new_journal.events.append({**e, **dict(offset=new_offset)})
            new_journal.data += journal.body.read()
        if not new_journal.is_empty:
            new_journal.upload()
            self._index_journal(new_journal)
        for o in objects_to_delete:
            o.delete()
        return new_journal

    def replay(self, from_date: datetime=None, to_date: datetime=None) -> typing.Iterator[Event]:
        search_range = DateRange(from_date, to_date)
        for journal_id in self.list_journals(from_date, to_date):
            print("replaying from journal", journal_id)
            journal = self._Journal.from_id(journal_id)
            for item in journal.events:
                event_date = datetime_from_timestamp(item['timestamp'])
                if event_date in search_range:
                    yield Event(item['event_id'], event_date, journal.body.read(item['size']))
                elif event_date in search_range.future:
                    break

    def _journal_for_event(self, event_id: str) -> JournalID:
        journal_id = self._KeyIndex.get(event_id)
        if journal_id is None:
            raise FlashFloodEventNotFound(f"journal not found for {event_id}")
        else:
            return journal_id

    def event_exists(self, event_id: str) -> bool:
        return self._KeyIndex.get(event_id) is not None

    def get_event(self, event_id: str) -> Event:
        journal_id = self._journal_for_event(event_id)
        journal = self._Journal.from_id(journal_id)
        return journal.get_event(event_id)

    def _generate_presigned_url(self, journal_id: JournalID):
        key = f"{self._blobs_pfx}/{journal_id.blob_id}"
        client = self.s3.meta.client
        return client.generate_presigned_url(ClientMethod="get_object",
                                             Params=dict(Bucket=self.bucket.name, Key=key))

    def list_journals(self, from_date: datetime=None, to_date: datetime=None) -> typing.Iterator[JournalID]:
        search_range = DateRange(from_date, to_date)
        for journal_id in self._Journal.list():
            journal_range = DateRange(journal_id.start_date, journal_id.end_date)
            if journal_range in search_range:
                yield journal_id
            elif journal_id.start_date in search_range.future:
                break

    def list_event_streams(self,
                           from_date: datetime=None,
                           to_date: datetime=None) -> typing.Iterator[typing.Mapping[str, typing.Any]]:
        for journal_id in self.list_journals(from_date, to_date):
            event_stream = self._Journal.from_id(journal_id).manifest()
            event_stream['stream_url'] = self._generate_presigned_url(journal_id)
            yield event_stream

    def _destroy(self):
        with S3Deleter(self.bucket) as s3d:
            for item in self.bucket.objects.filter(Prefix=f"{self.root_prefix}/"):
                s3d.delete(item.key)

def replay_event_stream(event_stream: dict, from_date: datetime=None, to_date: datetime=None) -> typing.Iterator[Event]:
    search_range = DateRange(from_date, to_date)
    for event_info in event_stream['events']:
        if datetime_from_timestamp(event_info['timestamp']) in search_range:
            break
    byte_range = f"bytes={event_info['offset']}-{event_stream['size']-1}"
    resp = requests.get(event_stream['stream_url'], headers=dict(Range=byte_range), stream=True)
    resp.raise_for_status()
    for item in event_stream['events']:
        event_date = datetime_from_timestamp(item['timestamp'])
        if event_date in search_range:
            yield Event(item['event_id'], event_date, resp.raw.read(item['size']))
        elif event_date in search_range.future:
            break



import time
import typing
import datetime
from string import hexdigits
from concurrent.futures import ThreadPoolExecutor, as_completed
from contextlib import AbstractContextManager

import boto3

def datetime_to_timestamp(dt):
    return dt.strftime("%Y-%m-%dT%H%M%S.%fZ")

def datetime_from_timestamp(ts):
    return datetime.datetime.strptime(ts, "%Y-%m-%dT%H%M%S.%fZ")

def timestamp_now():
    return datetime_to_timestamp(datetime.datetime.utcnow())

class DateRange:
    def __init__(self, start: datetime.datetime=None, end: datetime.datetime=None):
        self.start = start or datetime.datetime.min
        self.end = end or datetime.datetime.max
        assert self.start <= self.end

    def overlaps(self, other):
        """
        Test if [self.start : self.end] overlaps with (other.start : other.end]
        """
        a = self
        b = other
        if b.start <= a.start and a.end <= b.end:
            return True
        elif a.start <= b.start and b.end <= a.end:
            return True
        elif a.start <= b.start and b.start < a.end:
            return True
        elif a.start <= b.end and b.end <= a.end:
            return True
        return False

    def contains(self, date: datetime.datetime):
        """
        Test if (self.start : self.end] contains date
        """
        return self.start < date and date <= self.end

    def __contains__(self, item: typing.Any):
        if isinstance(item, datetime.datetime):
            return self.contains(item)
        elif isinstance(item, type(self)):
            return item.overlaps(self)
        else:
            raise TypeError(f"expected datetime instance or {type(self)} instance")

    @property
    def past(self):
        if datetime.datetime.min == self.start:
            return _EmptyDateRange()
        else:
            return type(self)(datetime.datetime.max, self.start)

    @property
    def future(self):
        if datetime.datetime.max == self.end:
            return _EmptyDateRange()
        else:
            return type(self)(self.end, datetime.datetime.max)

class _EmptyDateRange(DateRange):
    """
    This class represents an empty date range
    """
    def overlaps(self, *args, **kwargs):
        return False

    def contains(self, *args, **kwargs):
        return False

_S3_BATCH_DELETE_MAX_KEYS = 1000

def delete_keys(bucket, keys, number_of_workers=4):
    def _delete(keys_to_delete):
        assert _S3_BATCH_DELETE_MAX_KEYS >= len(keys_to_delete)
        objects = [dict(Key=key) for key in keys_to_delete]
        bucket.delete_objects(Delete=dict(Objects=objects))

    if keys:
        with ThreadPoolExecutor(max_workers=number_of_workers) as e:
            futures = list()
            while keys:
                futures.append(e.submit(_delete, keys[:_S3_BATCH_DELETE_MAX_KEYS]))
                keys = keys[_S3_BATCH_DELETE_MAX_KEYS:]
            for f in as_completed(futures):
                f.result()

class S3Deleter(AbstractContextManager):
    def __init__(self, bucket, deletion_threshold=5 * _S3_BATCH_DELETE_MAX_KEYS, number_of_workers=4):
        self.bucket = bucket
        self.deletion_threshold = deletion_threshold
        self._keys = list()
        self.number_of_workers = number_of_workers

    def delete(self, key):
        self._keys.append(key)

    def __exit__(self, *args, **kwargs):
        delete_keys(self.bucket, self._keys)

def concurrent_listing(bucket, prefixes, number_of_workers=4):
    """
    Concurrently list objects from `bucket` for `prefixes`. Lexicographical ordering is lost.
    """
    assert not isinstance(prefixes, str)

    def _list(pfx):
        # TDOD: handle or limit unbounded listing into memory
        return [item for item in bucket.objects.filter(Prefix=pfx)]

    with ThreadPoolExecutor(max_workers=number_of_workers) as e:
        futures = [e.submit(_list, f"{pfx}") for pfx in prefixes]
        for f in as_completed(futures):
            for item in f.result():
                yield item



from uuid import uuid4
from enum import Enum
from functools import lru_cache

from flashflood.util import datetime_from_timestamp, timestamp_now


TOMBSTONE_SUFFIX = ".dead"


class JournalUpdateAction(Enum):
    UPDATE = "update"
    DELETE = "delete"


class JournalID(str):
    DELIMITER = "--"

    @classmethod
    def make(cls, start_timestamp: str, end_timestamp: str, version: str, blob_id: str):
        return cls(start_timestamp + cls.DELIMITER + end_timestamp + cls.DELIMITER + version + cls.DELIMITER + blob_id)

    @classmethod
    def from_key(cls, key):
        return cls(key.rsplit("/", 1)[1])

    @lru_cache()
    def _parts(self):
        start_timestamp, end_timestamp, version, blob_id = self.split(self.DELIMITER)
        return start_timestamp, end_timestamp, version, blob_id

    @property
    def start_date(self):
        return datetime_from_timestamp(self._parts()[0])

    @property
    def end_date(self):
        end_date = self._parts()[1]
        if "new" == end_date:
            return self.start_date
        else:
            return datetime_from_timestamp(end_date)

    @property
    def version(self) -> str:
        return self._parts()[2]

    @property
    def blob_id(self):
        return self._parts()[3]

    @property
    def range_prefix(self):
        return self.rsplit(self.DELIMITER, 2)[0]


class JournalUpdateID(str):
    """
    This defines the id used to compose the object key on storage for journal updates.
    """
    DELIMITER = "--"

    @classmethod
    def make(cls, journal_id: str, event_id: str, action: JournalUpdateAction):
        reverse_journal_id = journal_id[::-1]
        return cls(reverse_journal_id + cls.DELIMITER
                   + event_id + cls.DELIMITER
                   + timestamp_now() + cls.DELIMITER
                   + action.name)

    @classmethod
    def from_key(cls, key):
        return cls(key.rsplit("/", 1)[1])

    @lru_cache()
    def _parts(self):
        reverse_journal_id, event_id, timestamp, action_name = self.rsplit(JournalID.DELIMITER, 3)
        return JournalID(reverse_journal_id[::-1]), event_id, timestamp, JournalUpdateAction[action_name]

    @property
    def journal_id(self) -> JournalID:
        return self._parts()[0]

    @property
    def event_id(self) -> str:
        return self._parts()[1]

    @property
    def timestamp(self) -> str:
        return self._parts()[2]

    @property
    def action(self) -> JournalUpdateAction:
        return self._parts()[3]

    @staticmethod
    def prefix_for_journal(journal_id: JournalID):
        return journal_id[::-1]
